package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;
import java.util.Random;

import explo.model.Option;
import explo.model.Arm;
import explo.model.User;
import explo.model.Batch;
import explo.agent.RegLinSimple;


// $Id: CP_Add_RegLinSimpleCondi.java,v 1.5 2011-04-10 21:26:55 garivier Exp $

/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_Add_RegLinSimpleCondi implements Serializable {

	private static final long serialVersionUID = 1L;
	/*
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	private Random r;
    	private int round;
		private int nbFeaturesCON = 99;
		private int nbFeaturesNOM = 20;
        // private Rescale scale[]; To compute statistics on the continuous data [unused]
    	private RegLinSimple rls[][];
		private RegClassSimple rcs[][];
		private int startCondi = 50000;
		private int condiFeature = -1;
		private boolean beFool = false;
	
	/**
	 * Default constructor.
	 */
	public CP_Add_RegLinSimpleCondi() {
		r = new Random();
		round = 0;
                // scale = new Rescale[1nbFeaturesNOM];
		// for (int i=0; i<1nbFeaturesNOM; i++) scale[i] = new Rescale();
		rls = new RegLinSimple[7][nbFeaturesCON];
		for(int o=0; o<7; o++) for(int i=0; i<nbFeaturesCON; i++) rls[o][i] = new RegLinSimple();
		rcs = new RegClassSimple[7][nbFeaturesNOM];
		for(int o=0; o<7; o++) for(int i=0; i<nbFeaturesNOM; i++) rcs[o][i] = new RegClassSimple();
	}
	
	/**
	 * Copy constructor: copies all fields of the class.
	 * @param cp - original click predictor to copy from
	 */
	public CP_Add_RegLinSimpleCondi(CP_Add_RegLinSimpleCondi cp) {	
		// TODO adapt to the structure of your click predictor so that all fields are copied
		r = new Random(); round = cp.round;
		rls = new RegLinSimple[7][nbFeaturesCON];
		rcs = new RegClassSimple[7][nbFeaturesNOM];
		for(int o=0; o<7; o++){
		    for (int i=0; i<nbFeaturesCON; i++) rls[o][i] = new RegLinSimple((RegLinSimple)cp.rls[o][i]); 
		    for (int i=0; i<nbFeaturesNOM; i++) rcs[o][i] = new RegClassSimple((RegClassSimple)(cp.rcs[o][i]));
		}
	}
	
        /** Learns from feedback received from the environment.
	 * @param a - arm that was played
	 * @param reward that was observed for a
	 */
	public void learn(Arm a, Integer reward) {
		// TODO implement your own code here
		// Be careful not to use more than half of the total available memory (explo.control.Run needs to keep a copy of your ClickPredictor instance).
		// You may want to call the garbage collector (System.gc();) for accurate memory usage reporting in the logs, but it might be very expensive if called at each iteration...
		
		/*Runtime runtime = Runtime.getRuntime();
		long free = runtime.freeMemory();
		long total = runtime.totalMemory();
		long used = total - free;*/ 
		round++;
		int o;
		if (round>startCondi) o = val2int(a.u.getFeaturesNOM()[condiFeature]); else o=0;
		if (o>6) o=6; // temporaire !
		nbFeaturesCON = a.u.getFeaturesCON().length;
		for (int i=0; i<nbFeaturesCON; i++){
			Number x=a.u.getFeaturesCON()[i];
			if (x != null) {
				rls[o][i].addPoint(x.floatValue(), reward);
			}
		}
		nbFeaturesNOM = a.u.getFeaturesNOM().length;
		for (int i=0; i<nbFeaturesNOM; i++){
			String x=a.u.getFeaturesNOM()[i];
			if (x != null){
				rcs[o][i].addPoint(val2int(x), reward);
			}
		}		
		// System.out.println("added point" + round + " ok");
	}
	
	/**
	 * Returns the index of the Arm in the Batch that is most likely to yield a click (indices start at 0). The arm is to be played in the environment.
	 * @param b - batch of arms
	 * @return index of the chosen arm
	 */	
	public int choose(Batch b) {
		if (beFool) return(2);
		// TODO implement your own code here
		
		/*
		 * Currently implementing epsilon-greedy  strategy with "additive" predictors
		 */
		if (round==startCondi){
			int minSupportSize=50;
			for(int f=0; f<nbFeaturesNOM; f++){
				//if(rcs[0][f].supportSize()<minSupportSize){
				if(rcs[0][f].supportSize()<5){
					condiFeature = f;
					minSupportSize = rcs[0][f].supportSize();
					if (r.nextInt(2)==0) break;
				}
			}
			if (minSupportSize>6) beFool = true; //minSupportSize = 6;
			for(int i=1; i<7; i++){
				for(int f=0; f<nbFeaturesCON; f++) rls[i][f] = new RegLinSimple((RegLinSimple)rls[0][f]);
				for(int f=0; f<nbFeaturesNOM; f++) rcs[i][f] = new RegClassSimple((RegClassSimple)rcs[0][f]); 
			}
			System.out.println("starting conditional work!");
		}
				 
		double score[] = new double[6];
		int o;
		for (int a=0; a<b.size(); a++){
			if (round>startCondi) o = val2int(b.getArm(a).u.getFeaturesNOM()[condiFeature]); else o=0;
			if (o>6) o=6; // temporaire !
			score[a] = 0;		
			for (int f=0; f<nbFeaturesCON; f++){
				Number x = b.getArm(a).u.getFeaturesCON()[f];
				if (x != null) score[a] += 1.0*lisse(rls[o][f].predict(x.floatValue()));
				else score[a]+=0.0024;
//				System.out.println(" x : " +x+ ", score = " + score[i]);
			}
			for (int f=0; f<nbFeaturesNOM; f++){
				String x = b.getArm(a).u.getFeaturesNOM()[f];
				if (x != null) score[a] += 1.0*rcs[o][f].predict(val2int(x));
				else score[a]+=0.0024;
//				System.out.println(" x : " +x+ ", score = " + score[i]);
			}			
			//System.out.println(" " + a + " : " + score[a]);
		}
		int toPlay = -1;
		double bestScore = -1;
		for (int a=0; a<b.size(); a++) if (score[a]>bestScore || ((score[a]==bestScore) && r.nextFloat()<0.3)){
			bestScore = score[a]; 
			toPlay = a;
		}
		if (toPlay<0 || r.nextFloat()<500/(500+0.1*Math.pow(round, 1.5)))
	                toPlay = r.nextInt(6);
		//System.out.println("  \\_ " + toPlay);
	    return(toPlay);
	}

        /** Utility routines go here */
	private int val2int(String s){
		return Integer.parseInt(s.substring(3));
	}
	
	private double lisse(double x){
		return Math.min(1, Math.max(0, x));
	}
	
	private int option2int(Option o){
		switch(o){
			case A: return(0);
			case B: return(1);
			case C: return(2);
			case D: return(3);
			case E: return(4);
			case F: return(5);
		}
		System.out.println("unknown option!"); 
		return 5;
	}	
}
