package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;
import java.util.Random;

import explo.model.Option;
import explo.model.Arm;
import explo.model.User;
import explo.model.Batch;



/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_Add_UCB implements Serializable {

	private static final long serialVersionUID = 1L;
	/*
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	private Random r;
    	private int round;
        // private Rescale scale[]; To compute statistics on the continuous data [unused]
    	private RegLinSimple regCON[];
		private RegClassSimple regNOM[];
    	private int nbFeaturesCON;
		private int nbFeaturesNOM;
	
	/**
	 * Default constructor.
	 */
	public CP_Add_UCB() {
		r = new Random();
		round = 0;
		regCON = new RegLinSimple[120];
		for (int i=0; i<120; i++) regCON[i] = new RegLinSimple();
		regNOM = new RegClassSimple[120];
		for (int i=0; i<120; i++) regNOM[i] = new RegClassSimple();
	}
	
	/**
	 * Copy constructor: copies all fields of the class.
	 * @param cp - original click predictor to copy from
	 */
	public CP_Add_UCB(CP_Add_UCB cp) {	
		// TODO adapt to the structure of your click predictor so that all fields are copied
		round = cp.round;
		for (int i=0; i<120; i++){
			regCON[i] = new RegLinSimple(cp.regCON[i]);
			regNOM[i] = new RegClassSimple(cp.regNOM[i]);
		}
	}
	
        /** Learns from feedback received from the environment.
	 * @param a - arm that was played
	 * @param reward that was observed for a
	 */
	public void learn(Arm a, Integer reward) {
		// TODO implement your own code here
		// Be careful not to use more than half of the total available memory (explo.control.Run needs to keep a copy of your ClickPredictor instance).
		// You may want to call the garbage collector (System.gc();) for accurate memory usage reporting in the logs, but it might be very expensive if called at each iteration...
		
		/*Runtime runtime = Runtime.getRuntime();
		long free = runtime.freeMemory();
		long total = runtime.totalMemory();
		long used = total - free;*/ 
		round++;
		nbFeaturesCON = a.u.getFeaturesCON().length;
		for (int i=0; i<nbFeaturesCON; i++){
			Number x=a.u.getFeaturesCON()[i];
			if (x != null) {
			    // scale[i].addPoint(x.floatValue());
			    regCON[i].addPoint(x.floatValue(), reward);
			}
		}
		nbFeaturesNOM = a.u.getFeaturesNOM().length;
		for (int i=0; i<nbFeaturesNOM; i++){
			String x=a.u.getFeaturesNOM()[i];
			if (x != null) regNOM[i].addPoint(val2int(x), reward);
		}		
		// System.out.println("added point" + round + " ok");
	}
	
	/**
	 * Returns the index of the Arm in the Batch that is most likely to yield a click (indices start at 0). The arm is to be played in the environment.
	 * @param b - batch of arms
	 * @return index of the chosen arm
	 */	
	public int choose(Batch b) {
		
		// TODO implement your own code here
		
		/*
		 * Currently implementing epsilon-greedy  strategy with "additive" predictors
		 */
		double score[] = new double[6];
		for (int a=0; a<b.size(); a++){
			score[a] = 0;
			for (int f=0; f<nbFeaturesCON; f++){
				Number x = b.getArm(a).u.getFeaturesCON()[f];
				if (x != null) score[a] += regCON[f].optimisticPredict(x.floatValue()); else score[a]+=0.0024; // ajouter 1 ou 0.0024 si on ne voit pas le regresseur ?
//				System.out.println(" x : " +x+ ", score = " + score[i]);
			}
			for (int f=0; f<nbFeaturesNOM; f++){
				String x = b.getArm(a).u.getFeaturesNOM()[f];
				if (x != null) score[a] += regNOM[f].optimisticPredict(val2int(x)); else score[a]+=0.0024;// ajouter 1 ou 0.0024 si on ne voit pas le regresseur ?
//				System.out.println(" x : " +x+ ", score = " + score[i]);
			}			

	//		System.out.println(" " + a + " : " + score[a]);
		}
		
		double bestScore = -1;
		boolean [] isBest = {false, false,false, false,false, false};
		int nBest = 0;
		for (int a=0; a<b.size(); a++) if (score[a]>=bestScore){
			isBest[a]=true;
			if (score[a]>bestScore){
				nBest = 1;
				bestScore = score[a];
				for(int aa=0; aa<a; aa++) isBest[aa]=false;
			}
			else nBest++;
		}
		int toPlay = r.nextInt(b.size());
		int k = r.nextInt(nBest); // we choose the k th champion
		boolean notFound = true;
		
		
		for (int a=0; notFound && a<b.size(); a++)
			if (isBest[a] && (k--==0)) {toPlay = a; notFound = false;}

		System.out.println("  \\_ " + toPlay);
	    return(toPlay);
	}

        /** Utility routines go here */
	private int val2int(String s){
		return Integer.parseInt(s.substring(3));
	}
	
	private double lisse(double x){
		return Math.min(1, Math.max(0, x));
	}
}
