package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;
import java.util.Random;

import explo.model.Option;
import explo.model.Arm;
import explo.model.User;
import explo.model.Batch;



/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_Combine implements Serializable {

	private static final long serialVersionUID = 1L;
	/*
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	public int round;
        public SimplePredictor regCON[];
		public SimplePredictor regNOM[];
		public double[][] pred;
		public Combiner combiner;
		public int nbFeaturesCON = 99;
		public int nbFeaturesNOM = 20;
		public int M = 32, M0 = 16, M1 = 16;
			
	/**
	 * Default constructor.
	 */
	public CP_Combine() {
		round = 0;
		combiner = new EXP3();
		regCON = new PredRobustCdfLin[nbFeaturesCON];
		for (int i=0; i<nbFeaturesCON; i++) regCON[i] = new PredRobustCdfLin(M, M0, M1); //PredRobustCdfLin(M, M0, M1); //PredRobustLin(M);
		regNOM = new PredClassSimple[nbFeaturesNOM];
		for (int i=0; i<nbFeaturesNOM; i++) regNOM[i] = new PredClassSimple();
		pred = new double[6][nbFeaturesCON + nbFeaturesNOM];
	}
	
	public CP_Combine(CP_Combine cp){
		round = cp.round;
		regCON = new PredRobustCdfLin[nbFeaturesCON];
		regNOM = new PredClassSimple[nbFeaturesNOM];
		for (int i=0; i<nbFeaturesCON; i++) regCON[i] = new PredRobustCdfLin((PredRobustCdfLin)cp.regCON[i]);
		for (int i=0; i<nbFeaturesNOM; i++) regNOM[i] = new PredClassSimple((PredClassSimple)cp.regNOM[i]);
		combiner = new EXP3((EXP3)cp.combiner);
		pred = new double[6][nbFeaturesCON + nbFeaturesNOM];
	}
	
	public void learn(Arm a, Integer reward) {
//		System.out.println("entering  CP_Combine.learn(arm="+a+", reward = "+reward+")...");
		round++;
		for (int i=0; i<a.u.getFeaturesCON().length; i++){
			Number x=a.u.getFeaturesCON()[i];
			if (x != null) regCON[i].addPoint(x.floatValue(), reward);
			}
			//System.out.println("... finished CON, starting NOM...");
		for (int i=0; i<a.u.getFeaturesNOM().length; i++){
			String x=a.u.getFeaturesNOM()[i];
			if (x != null) regNOM[i].addPoint(val2int(x), reward);
		}		
		combiner.updateWeights(reward, round);
//		System.out.println("...leaving  CP_Combine.learn()");
	}
	
	public int choose(Batch b) {
	//	System.out.println("entering  CP_Combine.choose()...");
		for (int a=0; a<b.size(); a++){
			Number[] featuresCON = b.getArm(a).u.getFeaturesCON();
			String[] featuresNOM = b.getArm(a).u.getFeaturesNOM();
			for (int f=0; f<nbFeaturesCON; f++){ 
				Number x = featuresCON[f];
				if (x != null) pred[a][f] = regCON[f].optimisticPredict(x.floatValue()); else pred[a][f] = 0.0024; 
			}
			for (int f=0; f<nbFeaturesNOM; f++){ 
				String x = featuresNOM[f];
				if (x != null) pred[a][nbFeaturesCON+f] = regNOM[f].optimisticPredict(val2int(x)); else pred[a][nbFeaturesCON+f] = 0.0024;
			}			
		}
		int toPlay = combiner.aggregate(pred, b.size());
	//	Random r = new Random(); if (r.nextFloat()<50/(50+round)){toPlay = r.nextInt(b.size());}
		/*if (round % 1000==999){
			//((PredRobustCdfLin) regCON[0]).X[0].print();
			//((PredRobustCdfLin) regCON[0]).X[1].print();
			PredLinSimple p = (PredLinSimple) regCON[1];
			System.out.println("xb = "+p.sx / p.n + ", sxy/n = "+p.sxy/p.n);
		}*/
//		System.out.println("... leaving  CP_Combine.choose()");
	    return(toPlay);
	}

        /** Utility routines go here */
	private int val2int(String s){
		return Integer.parseInt(s.substring(3));
	}	
}
