package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;
import java.util.Random;

import explo.model.Option;
import explo.model.Arm;
import explo.model.User;
import explo.model.Batch;



/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_CombineOption implements Serializable {

	private static final long serialVersionUID = 1L;
	/*
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	public int round;
        public SimplePredictor regCON[][]; //le premier est commun a toutes les options
		public SimplePredictor regNOM[][];
		public double[][] pred;
		public Combiner combiner;
		public int nbFeaturesCON = 99;
		public int nbFeaturesNOM = 20;
		public int M0=12, M=4;
			
	/**
	 * Default constructor.
	 */
	public CP_CombineOption() {
		round = 0;
		combiner = new EXP3();
		regCON = new PredRobustLin[7][nbFeaturesCON];
		for (int o=0; o<7; o++) for (int i=0; i<nbFeaturesCON; i++) regCON[o][i] = new PredRobustLin(o==0?M0:M);
		regNOM = new PredClassSimple[7][nbFeaturesNOM];
		for (int o=0; o<7; o++)  for (int i=0; i<nbFeaturesNOM; i++) regNOM[o][i] = new PredClassSimple();
		pred = new double[6][2*(nbFeaturesCON + nbFeaturesNOM)];
	}
	
	public CP_CombineOption(CP_CombineOption cp){
		round = cp.round;
		regCON = new PredRobustLin[7][nbFeaturesCON];
		regNOM = new PredClassSimple[7][nbFeaturesNOM];
		for (int o=0; o<7; o++) for (int i=0; i<nbFeaturesCON; i++) regCON[o][i] = new PredRobustLin((PredRobustLin)cp.regCON[o][i]);
		for (int o=0; o<7; o++)  for (int i=0; i<nbFeaturesNOM; i++) regNOM[o][i] = new PredClassSimple((PredClassSimple)cp.regNOM[o][i]);
		combiner = new EXP3((EXP3)cp.combiner);
		pred = new double[6][2*(nbFeaturesCON + nbFeaturesNOM)];
	}
	
	public void learn(Arm a, Integer reward) {
		//System.out.println("entering  CP_Combine.learn(arm="+a+", reward = "+reward+")...");
		round++;
		int o = option2int(a.o);
		for (int i=0; i<a.u.getFeaturesCON().length; i++){
			Number x=a.u.getFeaturesCON()[i];
			if (x != null){
				regCON[0][i].addPoint(x.floatValue(), reward);
				regCON[1+o][i].addPoint(x.floatValue(), reward);
			}
		}
			//System.out.println("... finished CON, starting NOM...");
		for (int i=0; i<a.u.getFeaturesNOM().length; i++){
			String x=a.u.getFeaturesNOM()[i];
			if (x != null){
				regNOM[0][i].addPoint(val2int(x), reward);
				regNOM[1+o][i].addPoint(val2int(x), reward);
			}
		}		
		combiner.updateWeights(reward, round);
		//System.out.println("...leaving  CP_Combine.learn()");
	}
	
	public int choose(Batch b) {
		//System.out.println("entering  CP_Combine.choose()...");
		for (int a=0; a<b.size(); a++){			
			int o = option2int(b.getArm(a).o);
			Number[] featuresCON = b.getArm(a).u.getFeaturesCON();
			String[] featuresNOM = b.getArm(a).u.getFeaturesNOM();
//			System.out.println("treating arm "+a);
			for (int f=0; f<nbFeaturesCON; f++){ 
				Number x = featuresCON[f];
				if (x != null){
					pred[a][f] = regCON[0][f].optimisticPredict(x.floatValue());
					pred[a][nbFeaturesCON + nbFeaturesNOM + f] = regCON[1+o][f].optimisticPredict(x.floatValue());
				}
				else {pred[a][f] = 0.0024; pred[a][nbFeaturesCON + nbFeaturesNOM + f]=0.0024;}
			}
			for (int f=0; f<nbFeaturesNOM; f++){ 
				String x = featuresNOM[f];
				if (x != null){
					pred[a][nbFeaturesCON+f] = regNOM[0][f].optimisticPredict(val2int(x)); 
					pred[a][nbFeaturesCON + nbFeaturesNOM + nbFeaturesCON + f] = regNOM[1+o][f].optimisticPredict(val2int(x)); 
				}
				else {pred[a][nbFeaturesCON+f] = 0.0024; pred[a][nbFeaturesCON + nbFeaturesNOM + nbFeaturesCON + f] = 0.0024;}
			}			
		}
		int toPlay = combiner.aggregate(pred, b.size());
	//	Random r = new Random(); if (r.nextFloat()<50/(50+round)){toPlay = r.nextInt(b.size());}
		if (round % 10000==-1){
			((PredRobustLin) regCON[0][0]).S[0].print();
			((PredRobustLin) regCON[0][0]).S[1].print();
			((PredRobustLin) regCON[1][0]).S[0].print();
			((PredRobustLin) regCON[1][0]).S[1].print();		
		}
//		System.out.println("...  CP_Combine.choose()");
	    return(toPlay);
	}

        /** Utility routines go here */
	private int val2int(String s){
		return Integer.parseInt(s.substring(3));
	}
	private int option2int(Option o){
		switch(o){
			case A: return(0);
			case B: return(1);
			case C: return(2);
			case D: return(3);
			case E: return(4);
			case F: return(5);
		}
		System.out.println("unknown option!"); 
		return 5;
	}
}
