package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;
import java.util.Random;

import explo.model.Option;
import explo.model.Arm;
import explo.model.User;
import explo.model.Batch;
import explo.agent.LDAsimple;


// $Id: CP_NaiveBayes.java,v 1.4 2011-04-02 13:49:39 garivier Exp $

/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_NaiveBayes implements Serializable {

	private static final long serialVersionUID = 1L;
	/*
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	private Random r;
    	private int round;
    	private LDAsimple regCON[];
		private RegClassSimple regNOM[];
    	private int nbFeaturesCON;
		private int nbFeaturesNOM;
	
	/**
	 * Default constructor.
	 */
	public CP_NaiveBayes() {
		r = new Random();
		round = 0;
                // scale = new Rescale[120];
		// for (int i=0; i<120; i++) scale[i] = new Rescale();
		regCON = new LDAsimple[120];
		for (int i=0; i<120; i++) regCON[i] = new LDAsimple();
		regNOM = new RegClassSimple[120];
		for (int i=0; i<120; i++) regNOM[i] = new RegClassSimple();
	}
	
	/**
	 * Copy constructor: copies all fields of the class.
	 * @param cp - original click predictor to copy from
	 */
	public CP_NaiveBayes(CP_NaiveBayes cp) {
		round = cp.round;
		for (int i=0; i<120; i++){ regCON[i] = cp.regCON[i]; regNOM[i] = cp.regNOM[i];}
		// TODO adapt to the structure of your click predictor so that all fields are copied
	}
	
        /** Learns from feedback received from the environment.
	 * @param a - arm that was played
	 * @param reward that was observed for a
	 */
	public void learn(Arm a, Integer reward) {
		// TODO implement your own code here
		// Be careful not to use more than half of the total available memory (explo.control.Run needs to keep a copy of your ClickPredictor instance).
		// You may want to call the garbage collector (System.gc();) for accurate memory usage reporting in the logs, but it might be very expensive if called at each iteration...
		
		/*Runtime runtime = Runtime.getRuntime();
		long free = runtime.freeMemory();
		long total = runtime.totalMemory();
		long used = total - free;*/ 
		round++;
		nbFeaturesCON = a.u.getFeaturesCON().length;
		for (int i=0; i<nbFeaturesCON; i++){
			Number x=a.u.getFeaturesCON()[i];
			if (x != null) {
			    regCON[i].addPoint(x.floatValue(), reward);
			}
		}
		nbFeaturesNOM = a.u.getFeaturesNOM().length;
		for (int i=0; i<nbFeaturesNOM; i++){
			String x=a.u.getFeaturesNOM()[i];
			if (x != null) regNOM[i].addPoint(val2int(x), reward);
		}		
		// System.out.println("added point" + round + " ok");
	}
	
	/**
	 * Returns the index of the Arm in the Batch that is most likely to yield a click (indices start at 0). The arm is to be played in the environment.
	 * @param b - batch of arms
	 * @return index of the chosen arm
	 */	
	public int choose(Batch b) {
		
		// TODO implement your own code here
		
		/*
		 * Currently implementing epsilon-greedy  strategy with "additive" predictors
		 */
		double score[] = new double[6]; // inutile d'initialiser a log(p/1-p) vu qu'on compare les options
		int toPlay = -1;
		double bestScore = -1e10;
		for (int a=0; a<b.size(); a++){
			score[a] = 0;		
			for (int f=0; f<nbFeaturesCON; f++){
				Number x = b.getArm(a).u.getFeaturesCON()[f];
				if (x != null) score[a] += lisse(regCON[f].predict(x.floatValue())); 
				//System.out.println(" x : " +x+ ", score = " + score[a]);
			}
			for (int f=0; f<nbFeaturesNOM; f++){
				String x = b.getArm(a).u.getFeaturesNOM()[f];
				if (x != null) score[a] += regNOM[f].predict(val2int(x)); 
			}			
			System.out.println(" " + a + " : " + score[a]);
			if ((score[a]>bestScore) || ((score[a]==bestScore) && r.nextFloat()<0.3)){
				bestScore = score[a]; 
				toPlay = a;
			}	
		}
		if ((toPlay<-1e10) || (r.nextFloat()<100/(100+Math.pow(round, 0.7)))) {toPlay = r.nextInt(b.size());} 
		System.out.println("  \\_ " + toPlay);	
	    return(toPlay);
	}

        /** Utility routines go here */
	private int val2int(String s){
		return Integer.parseInt(s.substring(3));
	}
	
	private double lisse(double x){
		return Math.min(3, Math.max(-3, x));
	}
}
