package explo.agent;

import java.util.HashMap;
import java.io.Serializable;
import java.lang.Math;

import explo.model.Option;
import explo.model.Arm;
import explo.model.Batch;

// $Id: CP_UCB_Option.java,v 1.1 2011-03-28 16:25:50 cappe Exp $

/**
 * ClickPredictors are used to predict which arm in a given batch is most likely to yield a reward of 1. They typically implement a learning algorithm.
 * The class is Serializable so that a ClickPredictor instance can be saved in a file and its learnings can be reused later. 
 * @author Louis Dorard, University College London
 */
public class CP_UCB_Option implements Serializable {

	private static final long serialVersionUID = 1L;
	/**
	 * This can be used for storing training data. You may change the type of this object depending on your needs.
	 */
    	private HashMap<Arm,Integer> training;
    	private int round;
    	private int[] cumReward;
    	private int[] nbPlayed;
    	
	
	/**
	 * Default constructor.
	 */
	public CP_UCB_Option() {

		training = new HashMap<Arm, Integer>();
		round = 0;
		cumReward = new int[6];
		nbPlayed = new int[6];
		for (int i=0; i<6; i++){
		    cumReward[i] = 0;
		    nbPlayed[i] = 0;
		}
	}
	
	/**
	 * Copy constructor: copies all fields of the class.
	 * @param cp - original click predictor to copy from
	 */
	public CP_UCB_Option(CP_UCB_Option cp) {

		training = cp.training;
		
		// TODO adapt to the structure of your click predictor so that all fields are copied
	}
	
	/** Utility: recode arm A-F letters to integers */
    	public int arm2int(Option o){
		switch(o){
			case A: return(0);
			case B: return(1);
			case C: return(2);
			case D: return(3);
			case E: return(4);
			case F: return(5);
		}
		return -1;

    	}

	/**
	 * Learns from feedback received from the environment.
	 * @param a - arm that was played
	 * @param reward that was observed for a
	 */
	public void learn(Arm a, Integer reward) {
		
		// TODO implement your own code here
		// Be careful not to use more than half of the total available memory (explo.control.Run needs to keep a copy of your ClickPredictor instance).
		// You may want to call the garbage collector (System.gc();) for accurate memory usage reporting in the logs, but it might be very expensive if called at each iteration...
		
		/*
		 * Sample code:
		 * here we don't really "learn" but we just memorise the training examples
		 */
		Runtime runtime = Runtime.getRuntime();
		long free = runtime.freeMemory();
		long total = runtime.totalMemory();
		long used = total - free;
		training.put(a, reward);
		if (used > 0.9*total/2) { // making sure we're not using more than 90% of half the total memory: you may want to leave some breathing space for running your computations
			training.remove(training.keySet().iterator().next());
		}
		round++;
		int i=arm2int(a.o);
		cumReward[i] += reward;
		nbPlayed[i]++;

		String s = "";
		for (int j=0; j<6; j++) s += nbPlayed[j]+" ";
		System.out.println(s);

	}
	
	/**
	 * Returns the index of the Arm in the Batch that is most likely to yield a click (indices start at 0). The arm is to be played in the environment.
	 * @param b - batch of arms
	 * @return index of the chosen arm
	 */
	public int choose(Batch b) {
		
		// TODO implement your own code here
		
		/*
		 * Currently implementing UCB on the "Option" index
		 */
	    if (round<6) {return(round);}
	    double maxUcb =  -1;
	    int toPlay = -1;
	    for (int j=0; j<6; j++){
		double ucb = cumReward[j] + Math.sqrt(Math.log(round)/(2*nbPlayed[j]));
		if (ucb>maxUcb){ maxUcb = ucb; toPlay=j;}
	    }
	    return(toPlay);

	}

}
