package explo.agent;
import java.lang.Math;
import java.util.Random;
import explo.agent.SmartSample;
import Jama.*; // pour ca il faut faire : set CLASSPATH=%CLASSPATH%;..\..\..\lib\Jama-1.0.2.jar;


public class PredCdfLin implements SimplePredictor{
	public SmartSample S;
	public double[][] X;
	public int[] n = new int[2];
	public int nbFeatures = 3; 
	private int memSize;
	
	public PredCdfLin(int M, int ms){
		memSize = ms;
		S = new SmartSample(M);
		X = new double[2][memSize];
		for (int i=0; i<2; i++)	n[i] = 0;
	}
	
	public PredCdfLin(PredCdfLin cp){
		memSize = cp.memSize;
		S = new SmartSample(cp.S);
		X = new double[2][memSize];
		for (int i=0; i<2; i++){
			n[i] = cp.n[i];
			for(int j=0; j<memSize; j++) X[i][j] = cp.X[i][j];
		}
	}
	
	public void addPoint(double x, int y){
		//System.out.println("entering PredCdfLin.addPoint()...");
		
		S.addPoint(x);
		X[y][n[y] % memSize] = x; // assumes that y==0 or y==1
		n[y]++;
		
		//System.out.println("...leaving PredCdfLin.addPoint()");
	}

	
	public double predict(double x){
		int m=Math.min(n[0], memSize) + Math.min(n[1], memSize);
		Matrix R = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)n[0]) / Math.min(n[0], memSize));
		double w1 = Math.sqrt(((double)n[1]) / Math.min(n[1], memSize));
		int j=0;
		for (; (j<memSize) && (j<n[0]); j++){
			double[] feature = buildFeatures(S.cdf(X[0][j]));
			for (int k=0; k<nbFeatures; k++) R.set(j, k, w0*feature[k]);
			Y.set(j, 0, 0.0);
		}
		for (int i=0; (i<memSize) && (i<n[1]); i++){
			double[] feature = buildFeatures(S.cdf(X[1][i]));
			for (int k=0; k<nbFeatures; k++) R.set(j+i, k, w1*feature[k]);
			Y.set(j+i, 0, w1);			
		}
		
		Matrix K = R.transpose().times(R);
		if (K.cond()>1e10) return(0.0024);
		Matrix theta = K.solve(R.transpose().times(Y));
		double[][] features = {buildFeatures(S.cdf(x))};
		Matrix v = new Matrix(features, 1, nbFeatures);
		double ph = v.times(theta).get(0,0);
		return(ph);	
	}
	
	public double optimisticPredict(double x){
		int m=Math.min(n[0], memSize) + Math.min(n[1], memSize);
		Matrix R = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)n[0]) / Math.min(n[0], memSize));
		double w1 = Math.sqrt(((double)n[1]) / Math.min(n[1], memSize));
		int j=0;
		for (; (j<memSize) && (j<n[0]); j++){
			double[] feature = buildFeatures(S.cdf(X[0][j]));
			for (int k=0; k<nbFeatures; k++) R.set(j, k, w0*feature[k]);
			Y.set(j, 0, 0.0);
		}
		for (int i=0; (i<memSize) && (i<n[1]); i++){
			double[] feature = buildFeatures(S.cdf(X[1][i]));
			for (int k=0; k<nbFeatures; k++) R.set(j+i, k, w1*feature[k]);
			Y.set(j+i, 0, w1);			
		}
		
		Matrix K = R.transpose().times(R);
		if (K.cond()>1e10) return(0.0024);
		Matrix theta = K.solve(R.transpose().times(Y));
		double[][] features = {buildFeatures(S.cdf(x))};
		Matrix v = new Matrix(features, 1, nbFeatures);
		double ph = v.times(theta).get(0,0);
		
		Matrix Yh = R.times(theta);
		double sumRes2 = 0;
		for(j=0; j<m; j++) sumRes2 += (Y.get(j, 0)-Yh.get(j, 0))*(Y.get(j, 0)-Yh.get(j, 0));
		double s = Math.sqrt(sumRes2/(n[0]+n[1]));

		double ucb = ph + 1.96 * s * Math.sqrt(v.times(K.solve(v.transpose())).get(0,0));		
		
		// System.out.println("... leavint PredRobustLin.optimisticPredict()");
		return(Math.min(1, Math.max(0, ucb)));		
	}
	
	public double[] buildFeatures(double x){
	//	double[] features = {1,x,x*x};
		double[] features = {1, x-0.5, x*x};
	//  double[] features = {1, x, x>S[0].getQuantile(S[1].M/4)?1:0};
		return(features);
	}
	
}
