package explo.agent;
import java.lang.Math;
import java.util.Random;
import explo.agent.SmartSample;
import Jama.*; // pour ca il faut faire : set CLASSPATH=%CLASSPATH%;..\..\..\lib\Jama-1.0.2.jar;


public class PredRobustCdfLin implements SimplePredictor{
	public SmartSample S;
	public SmartSample[] X;
	public int nbFeatures = 2; 
	
	public PredRobustCdfLin(int M, int M0, int M1){
		S = new SmartSample(M);
		X = new SmartSample[2];
		X[0] = new SmartSample(M0);
		X[1] = new SmartSample(M1);
	}
	
	public PredRobustCdfLin(PredRobustCdfLin cp){
		S = new SmartSample(cp.S);
		X = new SmartSample[2];
		for (int i=0; i<2; i++)	X[i] = new SmartSample(cp.X[i]);
	}
	
	public void addPoint(double x, int y){
		//System.out.println("entering PredCdfLin.addPoint()...");
		
		S.addPoint(x);
		X[y].addPoint(x);
		
		//System.out.println("...leaving PredCdfLin.addPoint()");
	}

	
	public double predict(double x){
		int m = X[0].M-1 + X[1].M-1; 
		Matrix R = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)X[0].size) / X[0].M);
		double w1 = Math.sqrt(((double)X[1].size) / X[1].M);
		for (int j=0; j<X[0].M-1; j++){
			double[] feature = buildFeatures(S.cdf(X[0].getQuantile(j+1)));
			for (int k=0; k<nbFeatures; k++) R.set(j, k, w0*feature[k]);
			Y.set(j, 0, 0.0);
		}
		for (int j=0; j<X[1].M-1; j++){
			double[] feature = buildFeatures(S.cdf(X[1].getQuantile(j+1)));
			for (int k=0; k<nbFeatures; k++) R.set(X[0].M-1+j, k, w1*feature[k]);
			Y.set(X[0].M-1+j, 0, w1);			
		}
		
		Matrix K = R.transpose().times(R);
		if (K.det()==0 || K.cond()>1e10)  return(0.0024);
		Matrix theta = K.solve(R.transpose().times(Y));
		double[][] features = {buildFeatures(S.cdf(x))};
		Matrix v = new Matrix(features, 1, nbFeatures);
		double ph = v.times(theta).get(0,0);
		return(ph);	
	}
	
	public double optimisticPredict(double x){
//		System.out.println("entering  PredRobustCdfLin.optimisticPredict()...");

		int m = X[0].M-1 + X[1].M-1; 
		Matrix R = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)X[0].size) / X[0].M);
		double w1 = Math.sqrt(((double)X[1].size) / X[1].M);
		for (int j=0; j<X[0].M-1; j++){
			double[] feature = buildFeatures(S.cdf(X[0].getQuantile(j+1)));
			for (int k=0; k<nbFeatures; k++) R.set(j, k, w0*feature[k]);
			Y.set(j, 0, 0.0);
		}
		for (int j=0; j<X[1].M-1; j++){
			double[] feature = buildFeatures(S.cdf(X[1].getQuantile(j+1)));
			for (int k=0; k<nbFeatures; k++) R.set(X[0].M-1+j, k, w1*feature[k]);
			Y.set(X[0].M-1+j, 0, w1);			
		}
		
		Matrix K = R.transpose().times(R);
		if (K.det()==0 || K.cond()>1e10) return(0.0024);
//		K.print(3, 1); 	System.out.println("condition number is "+K.cond());
		Matrix theta = K.solve(R.transpose().times(Y));
		double[][] features = {buildFeatures(S.cdf(x))};
		Matrix v = new Matrix(features, 1, nbFeatures);
		double ph = v.times(theta).get(0,0);
		
		Matrix Yh = R.times(theta);
		double sumRes2 = 0;
		for(int j=0; j<m; j++) sumRes2 += (Y.get(j, 0)-Yh.get(j, 0))*(Y.get(j, 0)-Yh.get(j, 0));
		double s = Math.sqrt(sumRes2/(X[0].size+X[1].size));

		double ucb = ph + 1.96 * s * Math.sqrt(v.times(K.solve(v.transpose())).get(0,0));		
		
//		 System.out.println("... leaving PredRobustCdfLin.optimisticPredict()");
		return(Math.min(1, Math.max(0, ucb)));		
	}
	
	public double[] buildFeatures(double x){
		//double[] features = {1,x-0.5,x>0.5 ? 1.0:0.0};
		//double[] features = {1,x-0.5,Math.sin(Math.PI*x)};
	//	double[] features = {1, x, Math.cos(Math.PI*x)};
	   double[] features = {1,x};
	//	double[] features = {1, x<0.5?1.0:-1.0, x<0.5 ? (x<0.25?1.0:-1.0) : 0.0, x>=0.5 ? (x<0.75?1.0:-1.0) : 0.0};
	//	double[] features =  {1, Math.cos(Math.PI*x), Math.sin(Math.PI*x)};
	//  double[] features = {1, x, x>S[0].getQuantile(S[1].M/4)?1:0};
		return(features);
	}
	
}
