package explo.agent;
import java.lang.Math;
import java.util.Random;
import explo.agent.SmartSample;
import Jama.*; // pour ca il faut faire : set CLASSPATH=%CLASSPATH%;..\..\..\lib\Jama-1.0.2.jar;


public class PredRobustLin implements SimplePredictor{
	public SmartSample[] S = new SmartSample[2];
	public int nbFeatures = 2; 
	
	public PredRobustLin(int M){
		S[0] = new SmartSample(M);
		S[1] = new SmartSample(M);
	}
	
	public PredRobustLin(PredRobustLin cp){
		S[0] = new SmartSample(cp.S[0]);
		S[1] = new SmartSample(cp.S[1]);
	}
	
	public void addPoint(double x, int y){
		//System.out.println("entering PredRobustLinSimple.addPoint()...");
		S[y].addPoint(x); // assumes that y==0 or y==1
		//System.out.println("...leaving PredRobustLinSimple.addPoint()");
	}

	
// robust version
	public double predict(double x){
		if (S[0].size<4 || S[1].size<4) return(1);
		int m = S[0].M-1 + S[1].M-1;
		Matrix X = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)S[0].size )/ S[0].M);
		double w1 = Math.sqrt(((double)S[1].size )/ S[1].M);
		for (int i=0; i<S[0].M-1; i++){
			double[] features = buildFeatures(S[0].getQuantile(i+1));
			for (int j=0; j<nbFeatures; j++) X.set(i, j, w0*features[j]);
			Y.set(i, 0, 0.0);
		}
		for (int i=0; i<S[1].M-1; i++){
			double[] features = buildFeatures(S[1].getQuantile(i+1));
			for (int j=0; j<nbFeatures; j++) X.set(S[0].M-1+i, j, w1*features[j]);
			Y.set(S[0].M-1+i, 0, w1);
		}
		
		Matrix K = X.transpose().times(X);
		if (K.det()==0 || K.cond()>1e10)  return(1);
		double[][] features = {buildFeatures(x)};
		Matrix v = new Matrix(features, 1, nbFeatures);
		Matrix theta = K.solve(X.transpose().times(Y));
		double ph = v.times(theta).get(0,0);
		return(ph);		
	}
	
	public double optimisticPredict(double x){
		// System.out.println("entering PredRobustLin.optimisticPredict()...");
		if (S[0].size<4 || S[1].size<4) return(1);
		int m = S[0].M-1 + S[1].M-1;
		Matrix X = new Matrix(m, nbFeatures);
		Matrix Y = new Matrix(m, 1);
		double w0 = Math.sqrt(((double)S[0].size )/ S[0].M);
		double w1 = Math.sqrt(((double)S[1].size )/ S[1].M);
		for (int i=0; i<S[0].M-1; i++){
			double[] features = buildFeatures(S[0].getQuantile(i+1));
			for (int j=0; j<nbFeatures; j++) X.set(i, j, w0*features[j]);
			Y.set(i, 0, 0.0);
		}
		for (int i=0; i<S[1].M-1; i++){
			double[] features = buildFeatures(S[1].getQuantile(i+1));
			for (int j=0; j<nbFeatures; j++) X.set(S[0].M-1+i, j, w1*features[j]);
			Y.set(S[0].M-1+i, 0, w1);
		}
		//X.print(6, 1);
		// System.out.println("computing in  PredRobustLin.optimisticPredict()...");
		
		Matrix K = X.transpose().times(X);
		if (K.det()==0 || K.cond()>1e10)  return(1);
		double[][] features = {buildFeatures(x)};
		Matrix v = new Matrix(features, 1, nbFeatures);
		Matrix theta = K.solve(X.transpose().times(Y));
		double ph = v.times(theta).get(0,0);
		
		Matrix Yh = X.times(theta);
		double sumRes2 = 0;
		for(int i=0; i<m; i++) sumRes2 += (Y.get(i, 0)-Yh.get(i, 0))*(Y.get(i, 0)-Yh.get(i, 0));
		double s = Math.sqrt(sumRes2/(S[0].size + S[1].size -1));

		double ucb = ph + s * Math.sqrt(v.times(K.solve(v.transpose())).get(0,0));		
		
		// System.out.println("... leavint PredRobustLin.optimisticPredict()");
		return(Math.min(1, Math.max(0, ucb)));		
	}
	
	public double[] buildFeatures(double x){
	//	double[] features = {1,x,x*x};
		double[] features = {1,x};
	//  double[] features = {1, x, x>S[0].getQuantile(S[1].M/4)?1:0};
		return(features);
	}
	
}
