package explo.agent;
import java.lang.Math;
import java.util.Random;


public class RegClassSimple{
	public int n=0;
	public double[] nx;
	public double[] sy;
	private double eps = 1e-10;

	public RegClassSimple(){
		nx = new double[50];
		sy = new double[50];
		for (int i=0; i<50; i++){nx[i]=0; sy[i]=0;}
	}

	public RegClassSimple(RegClassSimple cp){
		n = cp.n;
		nx = new double[50];
		sy = new double[50];
		for (int i=0; i<50; i++){nx[i] = cp.nx[i]; sy[i] = cp.sy[i];}
	}
	
	public void addPoint(int x, double y){
		n++;
		int xx = Math.min(x, 49);
		nx[xx]++;
		sy[xx] += y;
	}
	
	public double score(int x){
		//  cf lda : renvoie log p(y=1|x)/p(y=0|x) ** pour p(y=1)=1/2 **
		double beta=0.5; // cste bayesienne
		int xx = Math.min(x, 49);
		if (nx[xx]>0){
			return(Math.log(beta+sy[xx])-Math.log(beta+nx[xx]-sy[xx]));
		}
		else return(0);		
	}
	
	public double predict(int x){
		int xx = Math.min(x, 49);
		if (nx[xx]>0){
			return(sy[xx]/nx[xx]);
		}
		else return(0.0024);
	}
	
	public double optimisticPredict(int x){// version optimiste
		int xx = Math.min(x, 49);
		if (nx[xx]>0){
			double p = sy[xx]/nx[xx];
			//double ucb = ucbkl(p, Math.log(1+n)/((double)nx[xx])); 
			double ucb = ucbkl(p, 1.0/((double)nx[xx])); 
			//double ucb = p+Math.sqrt(Math.log(1+n)/2.0*p/nx[xx]); 
			//double ucb = p+Math.sqrt(1.0*p/((double)(nx[xx]))); 
			return(ucb); //Math.min(1,ucb));
		}
		else return(1);
	}	
	
	public int supportSize(){
		int n=0;
		for(int i=0; i<50; i++) if (nx[i]!=0) n=i;
		return n;
	}

	
	private double kl(double p, double q){
		double x = Math.min(1-eps, Math.max(eps, p));
		double y = Math.min(1-eps, Math.max(eps, q));
		return(x*Math.log(x/y) + (1-x)*Math.log((1-x)/(1-y)));
	}

	private double ucbkl(double p, double d){
		double l=p, u=1; 
		for (int i=0; i<15; i++){
			double m = (l+u)/2.0;
			if (kl(p, m)<d) l=m;
			else u=m;
		}
		return(u);
	}


 //	just for testing : comment first line 'package explo.agent', uncomment this, and compile with javac
/*	public static void main(String[] args){
		System.out.println("coucou");
		Random r = new Random();
		RegClassSimple rcs = new RegClassSimple();
		int n = 10000;
		for (int i=0; i<n; i++){
			double x = r.nextInt(3);
			double y = r.nextFloat()<(x-1.0)*(x-1.0)/2+x/5.0 ? 1.0:0.0;
			rcs.addPoint((int)x,y);
		}
		
		int x = 0;
		double yh = rcs.predict(x);
		double yucb = rcs.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);
		x = 1;
		yh = rcs.predict(x);
		yucb = rcs.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);	
		x = 2;
		yh = rcs.predict(x);
		yucb = rcs.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);		
	}	
*/
}
