package explo.agent;
import java.lang.Math;
import java.util.Random;
import Jama.*; // pour ca il faut faire : set CLASSPATH=%CLASSPATH%;..\..\..\lib\Jama-1.0.2.jar;


public class RegLin{
	public Matrix K,L;
	public int nbFeatures = 2;
	public int n = 0;
	
	public RegLin(){
		K = new Matrix(nbFeatures, nbFeatures);
		L = new Matrix(nbFeatures, 1);
	}
	
	public RegLin(RegLin cp){
		K = cp.K.copy(); L =cp.L.copy();
		nbFeatures = cp.nbFeatures; n = cp.n;
	}
	
	public Matrix buildFeatures(double x){
	//	double[][] features = {{1,x,x*x}};
		double[][] features = {{1,x}};
		Matrix v = new Matrix(features, 1, nbFeatures);
		return(v);
	}
	
	public void addPoint(double x, double y){
		n++;
		Matrix v = buildFeatures(x);
		K.plusEquals(v.transpose().times(v));
		L.plusEquals(v.transpose().times(y));
	}
	
	public double optimisticPredict(double x){
		if (K.det() != 0){ // test that K is inversible?
			Matrix v = buildFeatures(x);
			double ph = v.times(K.solve(L)).get(0,0);
			double sig2 = v.times(K.solve(v.transpose())).get(0,0); 
			if(sig2<0) return(1); // attention instabilite numerique ici
			// double ucb = ph + Math.sqrt(Math.log(1+n)/2.0 * sig2);
			double ucb = ph + Math.sqrt(Math.log(1+n)/2.0 * sig2);
			if (Double.isNaN(ph)||Double.isNaN(ucb)) System.out.println("Warning : NaN ph = "+ph+", sig2 = "+sig2+", ucb = "+ucb);
			return(Math.min(1, Math.max(0, ucb)));
		}
		else return(1);
	}
	
	public double predict(double x){
		if (K.det() != 0){ // test that K is inversible?
			Matrix v = buildFeatures(x);
			return(v.times(K.solve(L)).get(0,0));
		}
		else return(0.0024);
	}

	
 //	just for testing : comment first line 'package explo.agent', uncomment this, and compile with javac
/*	public static void main(String[] args){
		System.out.println("RegLin");
		Random r = new Random();
		RegLin rl = new RegLin();
		int n = 10000;
		for (int i=0; i<n; i++){
			double x = r.nextFloat();
			double y = r.nextFloat()<(x-0.25)*(x-0.25)+1-x?1.0:0.0;
			rl.addPoint(x,y);
		}
		double x = 0.2;
		double yh = rl.predict(x);
		double yucb = rl.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);
		x = 0.5;
		yh = rl.predict(x);
		yucb = rl.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);	
		x = 0.75;
		yh = rl.predict(x);
		yucb = rl.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);
	}
*/	
}
