package explo.agent;
import java.lang.Math;

/** Class for doing online one-dimensional logistic regression based on
 * Online Newton alg. of "Logarithmic Regret Algorithms for Online Convex
 * Optimization", Hazan, Kalai et al.
 * Commented are modifications to use the logistic Hessian formulas (see
 * rech/theses/filippi/textes/ParametBandit/bin/logreg/online_logreg_step.m for Matlab vervion) */
public class RegLogSimple{
	public double n=0;
	public double theta0=0;
	public double theta1=0;
	public double beta=1;
	public double g0=0;
	public double g1=0;
	public double H00=0;
	public double H01=0;
	public double H11=0;

	
	public RegLogSimple(){}
	
	public void addPoint(double x, double y){
		double p;
		double g0_cur;
		double g1_cur;
		double det;

 		n++;
		// Prediction
		p = logitinv(x);
		// Current gradient
		g0_cur = y-p;
		g1_cur = (y-p)*x;
		// Smoothed gradient
		/* g0 = g0 + (1-p)*p*(theta0+theta1*x) + g0_cur/beta;
		   g1 = g1 + (1-p)*p*x*(theta0+theta1*x) + g1_cur/beta; */
		g0 += g0_cur*(g0_cur*theta0+g1_cur*theta1) + g0_cur/beta;
		g1 += g1_cur*(g0_cur*theta0+g1_cur*theta1) + g1_cur/beta;
		// Smoothed Hessian
		/* H00 += (1-p)*p;
		   H01 += (1-p)*p*x;
		   H11 += (1-p)*p*x*x; */
		H00 += g0_cur*g0_cur;
		H01 += g0_cur*g1_cur;
		H11 += g1_cur*g1_cur;
		// Matrix inversion (if feasible)
		det = H00*H11-H01*H01;
		if (det > 0) {
			theta0 = (H11 - x*H01)/det;
			theta1 = (-H01 + H00*x)/det;
		}
	}
	
	public double logitinv(double x){
	    return(1/(1+Math.exp(-theta0+theta1*x)));
	}
	/** Prediction with log-prob */
	public double predict(double x){
	    return(-Math.log(1+Math.exp(-theta0+theta1*x)));
	}
}
