package explo.agent;
import java.lang.Math;
import java.util.Random;
import java.util.*;

public class RegQuantiles{
	public int m;
	public SmartSample S;
	public ArrayList<Double> S1;
	public int n = 0;
	private double eps = 1e-10;	
	
	public RegQuantiles(int mm){
		m = mm;
		S = new SmartSample(m);
		S1 = new ArrayList<Double>();
	}
	
	public RegQuantiles(RegQuantiles cp){
		m = cp.m;
		S = cp.S;
		S1 = cp.S1;
		n = cp.n;	
	}
		
	public void addPoint(double x, int y){
		n++;
		S.addPoint(x);
		if (y==1){
			S1.add(x);
			Collections.sort(S1);
		}
	}
	
	public double optimisticPredict(double x){
		int pos = S.pos(x);
		pos = Math.min(m-1, Math.max(0, pos));
		int i=0;
		while (i<S1.size() && S1.get(i++)<S.q[pos]) ;
		int N = 0;
		while (i<S1.size() && S1.get(i++)<S.q[pos+1]) N++;
		double nj = (double)(S.size) / m;
		double ph = (double)(N) / nj;
		//double ucb = ucbkl(ph, Math.log(1+n)/nj);
		double ucb = ucbkl(ph, 1.0/nj);
		return(ucb);	
	}
	
	public double predict(double x){
		int pos = S.pos(x);
		pos = Math.min(m-1, Math.max(0, pos));
		int i=0;
		while (i<S1.size() && S1.get(i++)<S.q[pos]) ;
		int N = 0;
		while (i<S1.size() && S1.get(i++)<S.q[pos+1]) N++;
		double nj = (double)(S.size) / m;
		double ph = (double)(N) / nj;
		return(ph);
	}

	private double kl(double p, double q){
		double x = Math.min(1-eps, Math.max(eps, p));
		double y = Math.min(1-eps, Math.max(eps, q));
		return(x*Math.log(x/y) + (1-x)*Math.log((1-x)/(1-y)));
	}

	private double ucbkl(double p, double d){
		double l=p, u=1; 
		for (int i=0; i<15; i++){
			double m = (l+u)/2.0;
			if (kl(p, m)<d) l=m;
			else u=m;
		}
		return(u);
	}	
	
 //	just for testing : comment first line 'package explo.agent', uncomment this, and compile with javac
/*	public static void main(String[] args){
		System.out.println("RegQuantiles");
		Random r = new Random();
		RegQuantiles rq = new RegQuantiles(10);
		int n = 10000;
		for (int i=0; i<n; i++){
			double x = r.nextFloat();
			int y = r.nextFloat()<(x-0.25)*(x-0.25)+1-x?1:0;
			rq.addPoint(x,y);
		}
		double x = 0.2;
		double yh = rq.predict(x);
		double yucb = rq.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);
		x = 0.5;
		yh = rq.predict(x);
		yucb = rq.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);	
		x = 0.75;
		yh = rq.predict(x);
		yucb = rq.optimisticPredict(x);
		System.out.println(x + " : " + yh + ", ucb = " + yucb);
	}
	*/
}
