package explo.agent;
import java.lang.Math;
import java.util.Random;
import explo.agent.SmartSample;


public class RobustLDA{
	public SmartSample[] S = new SmartSample[2];
	public double p;
	public double[] med = new double[2];
	public double iqr;
	
	public RobustLDA(int M){
		S[0] = new SmartSample(M);
		S[1] = new SmartSample(M);
	}
	
	public RobustLDA(RobustLDA cp){
		p = cp.p;
		iqr = cp.iqr;
		int M = cp.S[0].M;
		S[0] = new SmartSample(M);
		S[1] = new SmartSample(M);		
		for(int i=0; i<2; i++){S[i] = cp.S[i]; med[i] = cp.med[i];}
	}
	
	public void addPoint(double x, int y){
		S[y].addPoint(x); // assumes that y==0 or y==1
	}
	
	public void estimateParameters(){
		p = (S[1].size+0.24)/(S[1].size+S[0].size+100);
		double[] iq = new double[2];
		for (int i=0; i<2; i++){
			med[i] = S[i].getMedian();
			iq[i] = S[i].getIQR();
		}
		iqr = (S[0].size*iq[0]+S[1].size*iq[1]) /(S[0].size+S[1].size);
	}
	
	public double score(double x){
		estimateParameters();
		return(2*Math.log(2)*(Math.abs(x-med[0])/iqr - Math.abs(x-med[1])/iqr));
	}
	
	public double predict(double x){
		double el = Math.exp(Math.min(500,-score(x)));
		return (p/(1-p)/(1+el));
	}
/*
	public static void main(String[] args){
		System.out.println("robustLDA");
		Random r = new Random();
		RobustLDA lda = new RobustLDA(100);
		double p=0.1;
		double[] mu = {-1, 1};
		double[] sig2 = {1, 1};
		int n=10000;
		for(int i=0; i<n; i++){
			int y = r.nextFloat()<p?1:0;
			double x = mu[y]+Math.sqrt(sig2[y])*r.nextGaussian();
			lda.addPoint(x, y);
			//S.print();
		}
		lda.S[0].print();
		lda.S[1].print();
		double x = -2; System.out.println("x="+x+", score = "+lda.score(x)+", P(1)="+lda.predict(x));
		x = -1; System.out.println("x="+x+", score = "+lda.score(x)+", P(1)="+lda.predict(x));
		x = 0; System.out.println("x="+x+", score = "+lda.score(x)+", P(1)="+lda.predict(x));
		x = 1; System.out.println("x="+x+", score = "+lda.score(x)+", P(1)="+lda.predict(x));
		x = 2; System.out.println("x="+x+", score = "+lda.score(x)+", P(1)="+lda.predict(x));
	}
	*/
}