package explo.agent;

import java.lang.Math;
import java.util.Random;
import java.util.Arrays;


public class SmartSample{
	public int size, M, bufferSize, bufferFactor=100;
	public double[] q;
	public double[] qm;
	public double[] buffer;
	
	public SmartSample(int MM){
		M = MM; bufferSize = bufferFactor*M+1;
		size = 0;
		q = new double[M+1];
		qm = new double[M+1];
		buffer = new double[bufferSize];
	}
	
	public SmartSample(SmartSample cp){
		size = cp.size; M = cp.M; bufferSize=cp.bufferSize;
		q = new double[M+1]; qm = new double[M+1];
		buffer = new double[bufferSize];
		for(int i=0; i<M+1; i++){
			q[i] = cp.q[i];
			qm[i] = cp.qm[i];
		}
		for(int i=0; i<bufferSize; i++) buffer[i]=cp.buffer[i];
	}	
	
	public void addPoint(double x){
		//System.out.println("entering SmartSample.addpoint(x="+x+")...");
		if (++size<=bufferSize){
			buffer[size-1] = x;
			Arrays.sort(buffer, 0, size);
			if (size==bufferSize){
				for (int i=0; i<=M; i++){
					q[i] = buffer[bufferFactor*i];
					qm[i] = q[i];
				}
				bufferSize = 0; 
				buffer = new double[0];
			}
		}
		else{
			int p = pos(x); 
			if (p == -1){
				q[0] = x; qm[0] = x;
			}
			else if (p == M){
				q[M] = x; qm[M] = x;
			}
			else{
				double step = as(size);
			// rule 1: local uniform step
				if (p>0) q[p] += step;
				if (p<M-1) q[p+1] -= step; 
			// rule 2: global steps, seems to be better
			/*  for (int j=1; j<=p; j++) q[j] += 1.0/(M-j) * step;
				for (int j=p+1; j<M; j++) q[j] -= 1.0/j * step; */
			/* // rule 3: local adaptive step NOT STABLE: QUANTILES ARE NOT THE ONLY FIXPOINT! 
				if (p>0) q[p] += (q[p+1]-q[p-1])/2.0 * step;
				if (p<M-1) q[p+1] -= (q[p+2]-q[p])/2.0 * step; */
				
				for (int j=1; j<M; j++) qm[j] = (1-1.0/size)*qm[j] + 1.0/size * q[j];
			}
		}
		//System.out.println("leaving SmartSample.addpoint");
	}
	
	public int pos(double x){
	// returns the index of the smallest l such that qm[l]<=x<qm[l+1]
	// returns -1 if x<qm[0], M if x>qm[M]
		//System.out.println("entering SmartSample.pos(x="+x+")...");
		if (size == 0) return(-1); 
		else if (size<bufferSize){
			int l = -1;
			int u = size+1;
			while (u-l>1){
				int m = (int)((u+l)/2);
				if (x<buffer[m]) u = m;	else l = m;
			}
			return(l/bufferFactor);		
		}
		else{
			int l = -1;
			int u = M+1;
			while (u-l>1){
				int m = (int)((u+l)/2);
				if (x<qm[m]) u = m; else l = m;
			}
			return(l);
		}
	}
	
	public double cdf(double x){
//		System.out.println("entering SmartSample.cdf(x="+x+")...");
		if (size == 0) return(0.5); 
		else if (size<bufferSize){
			if (x<buffer[0]) return(0);
			else if (x>buffer[size-1]) return(1);
			else{
				
				int l = 0;
				int u = size+1;
				while (u-l>1){
					int m = (int)((u+l)/2);
					if (x<buffer[m]) u = m;	else l = m;
				}
				return(((double)l)/((double)bufferSize));
			}	
		}
		else{
			int p = pos(x);
			if (p==-1) return(0);
			else if (p==M) return(1);
			else return (((double)p + (x-qm[p])/(qm[p+1]-qm[p])) / ((double)M));
		}
	}

	public double getQuantile(int i){
//		System.out.println("entering SmartSample.getQuantile(i=" + i+")...");
		if (size<bufferSize) return buffer[(int)(size*i/M)];
		else return qm[i];
	}

	public double getMedian(){
		if (size<bufferSize){
			return(buffer[size/2]);
		}
		else{
			if (M % 2==0) return(qm[M/2]);
			else return((qm[(M-1)/2] + qm[(M+1)/2])/2);
		}
	}
	
	public double getIQR(){
		if (size<bufferSize){
			return(buffer[(int)(3*size/4)]-buffer[(int)(size/4)]);			
		}
		else{
			return(qm[(int)(3*M/4)] - qm[(int)(M/4)]);
		}
	}
	
	public double estimateMean(){
		if (size<bufferSize){
			double s=0;
			for(int i=1; i<size-1; i++){ // forget extreme quantiles
				s += buffer[i];
			}
			return(s/(size-2));
		}
		else{
			double s=0;
			for(int i=1; i<M; i++){ // forget extreme quantiles
				s += qm[i];
			}
			return(s/(M-1));
		}
	}
		
	public double estimateVariance(){
		if (size<bufferSize){	
			double s=0, s2=0;
			for(int i=1; i<size-1;  i++){ // forget extreme quantiles
				s += buffer[i];
				s2 += buffer[i]*buffer[i];
			}
			return (s2-s*s/(size-1))/Math.max(size-2, 1);
		}
		else{
			double s=0, s2=0;
			for(int i=1; i<M; i++){ // forget extreme quantiles
				s += qm[i];
				s2 += qm[i]*qm[i];
			}
			return (s2-s*s/(M-1))/Math.max(M-2, 1);		
		}
	}
	
	public double estimateVarianceIQR(){//specific to gaussian distribution
		double std = getIQR()/1.348980;
		return(Math.max(std*std, 1e-10));
	}
	
	private double as(int n){
		return 1.0/Math.pow(n, 0.5);
	}
	
	public void print(){
		System.out.print("sample of size "+size);
		System.out.println(" -- median = "+getMedian()+", IQR = "+getIQR());
		for(int i=0; i<M+1; i++) System.out.printf("%1.4f ", qm[i]);
		System.out.println(".");
	}
/*
	public static void main(String[] args){
		System.out.println("SmartSample");
		Random r = new Random();
		int n=100000;
		SmartSample S = new SmartSample(100);
		for(int i=0; i<n; i++){
			S.addPoint(r.nextGaussian());
			//S.print();
		}
		S.print();
		double m = S.getMedian();
		double v = S.estimateVarianceGaussienne();
		System.out.println("moyenne : "+ m+", variance :"+v);
	}
	*/
}
